/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	Copyright 1998-2002 Anton Vinokurov <anton@netams.com>
***	Copyright 2002-2008 NeTAMS Development Team
***	This code is GPL v3
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: flow.c 1628 2013-10-28 19:08:22Z anton $ */

#include "config.h"

struct ipv4_key* IPv4GetKey(struct ip *ip, struct ipv4_key* ipv4_flow_key){

	ipv4_flow_key->ipv4_info.ip_p    = ip->ip_p;
	// ipv4_flow_key->ipv4_info.ip_tos  = ip->ip_tos;
	ipv4_flow_key->ipv4_info.ip_src  = ip->ip_src;
	ipv4_flow_key->ipv4_info.ip_dst  = ip->ip_dst;

	if(ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP) {
		struct tcphdr       *th = (struct tcphdr *)((u_char *)ip + ip->ip_hl*4);
		ipv4_flow_key->tcp_info.src_port	= th->th_sport;
		ipv4_flow_key->tcp_info.dst_port	= th->th_dport;
	} else {
		ipv4_flow_key->tcp_info.src_port	= 0;
		ipv4_flow_key->tcp_info.dst_port	= 0;
	}
	return ipv4_flow_key;
}

//////////////////////////////////////////////////////////////////////////

flow::flow(){
	next=NULL;
	ex_next=NULL;
	ex_packets=0; ex_time=0;
	st_time=time(NULL);
	packets=0;
	bytes=0;
	authflags=0;
	flushflag=0;
	agg_count=0;
	memset(&key, '\0', sizeof(struct ipv4_key));
	hash=0;
	status=FLOW_STATUS_NEW;

#ifdef URL_CHECK
	url_len=0;
	url_value=NULL;
#endif

#ifdef ENABLE_DPI
	dpi_protocol=0;
	ipoque_flow = (struct ipoque_flow_struct *)malloc(size_flow_struct);
#endif

}

flow::~flow(){
mem_c_total_free++;
mem_total_free+=sizeof(flow);
#ifdef URL_CHECK
	if (url_value!=NULL) {
		mem_c_url_free++;
		mem_url_free+=url_len;
		free(url_value);
	}
#endif
#ifdef ENABLE_DPI
	free(ipoque_flow);
#endif
}

//////////////////////////////////////////////////////////////////////////

flowcache::flowcache(struct ds_pcap_info *p){
	entries=0;
	total_entries=0;
	flushed_entries=0;
	routed_entries=0;
	ex_root=NULL;
	pcap_info=p;
	cache = (flow**)malloc(sizeof(flow*)* FLOW_CACHE_SIZE);
	for (int i=0; i<FLOW_CACHE_SIZE; i++) cache[i]=NULL;
}

flow *flowcache::match(struct ipv4_key *key, unsigned long packets, unsigned long long bytes){
	unsigned short hash=FLOW_HASH((u_char*)key);

	flow *f = cache[hash];
	char p1[32], p2[32];
	
	// look for the flow and update counters
	for (; f!=NULL; f=f->next){
		if (!memcmp(&f->key, key, sizeof(struct ipv4_key))) {
			f->bytes += bytes;
			f->packets += packets;
			return f;
		}
	}

	// there was no matched flow, create one.
	f = new flow(); mem_total_alloc+=sizeof(flow); mem_c_total_alloc++;
	memcpy(&f->key, key, sizeof(struct ipv4_key));
	f->authflags=0; // change this
	f->hash=hash;
	entries++;
	total_entries++;

	f->next = cache[hash];
	cache[hash] = f;
	// update expiresearch (ex_) references
	f->ex_next=ex_root;
	ex_root=f;

	f->bytes += bytes;
	f->packets += packets;
// 	n4debug("CREATE flow %p [%u %s:%u -> %s:%u]\n", f, key->ipv4_info.ip_p, inet_ntop(AF_INET, &key->ipv4_info.ip_src, p1, 32), ntohs(key->tcp_info.src_port), inet_ntop(AF_INET, &key->ipv4_info.ip_dst, p2, 32), ntohs(key->tcp_info.dst_port));
	(pcap_info->submit_handler)(f); // check auth

	return f;
}

void flowcache::expire(){
	time_t now = time(NULL);
	flow *f  = ex_root;
	int flushflag=0;
	unsigned count_entries=0;
	for (; f!=NULL; f=f->ex_next){
		count_entries++; if (count_entries > (entries+10) ) {
			syslog(LOG_INFO, "NETAMS4 remote agent: flowcache::expire loop detected\n");
			break; // safety
		}
		if ((now - f->st_time)> 5*pcap_info->flow_expiretime) { // too old flow, default=10 minutes : expire (absolute timeout)
			n4debug("FFLUSH flow %p [start: %lu now: %lu, lastpacket: %lu delta=%lu]\n", f, f->st_time, now, f->ex_time, ((now - f->st_time)));
			f->flushflag=1;
			flushflag=1;
		} else {
			if (f->ex_packets < f->packets) { // since last expiresearch, there was some traffic
				f->ex_time=now; // update last activity timestamp
				f->ex_packets=f->packets; // update packet count
			} else { // there was no traffic since expiresearch
				if (now - f->ex_time >= pcap_info->flow_expiretime) { // there was no traffic for active_timeout seconds (120) - push this flow out
					f->flushflag=1;
					flushflag=1;
				}
			}
		}
	}
	if (flushflag) flush();
}

void flowcache::flush(){
	flow *f1, *f2 = NULL;
	flow *purge_root=NULL;
	int i1=0, i2=0, p=0;
	if (ex_root==NULL) return;
	for (f1=ex_root; f1!=NULL;){
		i1++;
		if (f1->flushflag == 1) {
//			n4debug("FLUSH flow %p, prev=%p, next=%p , ex_next=%p, pkts=%lu\n", f1, f2, f1->next, f1->ex_next, f1->packets);
			// take out from ex_ list
			if (f1==ex_root) {
				ex_root=f1->ex_next;
				f2=ex_root;
			} else
				f2->ex_next=f1->ex_next;
			// take out of hash list
			flow *h1, *h2 = NULL;
			for (h1=cache[f1->hash]; h1!=NULL; h2=h1, h1=h1->next){
				if (h1==f1) {
					if (h1==cache[f1->hash])
						cache[f1->hash]=h1->next;
					else
						h2->next=h1->next;
					break;
				}
			}
			//return back to processor
			f1->status = FLOW_STATUS_EXPIRED;
			flushed_entries++;
			if (pcap_info->optimization==0 || f1->bytes>PURGE_THRESHOLD || !(f1->key.ipv4_info.ip_p==6 || f1->key.ipv4_info.ip_p==17) ) {
				(pcap_info->submit_handler)(f1);
				routed_entries++;
				delete f1;
			}
			else { // aggregation, move to purge cache
				f1->next=purge_root;
				purge_root=f1;
				p++;
			}
			entries--;
			i2++;
			f1=f2;
		} else {
			f2=f1;
			f1=f1->ex_next;
		}
	}
	n4debug("FLUSH %d of %d flows, PURGE CACHE: %d\n", i2, i1, p);
	if (p>0) {
		char p1[32], p2[32];

		for (f1=purge_root; f1!=NULL; f1=f1->next) {
			ipv4_key *k = &f1->key;
			unsigned short hash=FLOW_HASH((u_char*)k);
			unsigned long z = FLOW_SHORT_HASH(k);
			n4debug("\tDUMP %lu %04x %p [%u %s:%u -> %s:%u]%s %lu %llu \n", z, hash, f1, k->ipv4_info.ip_p, inet_ntop(AF_INET, &k->ipv4_info.ip_src, p1, 32), ntohs(k->tcp_info.src_port), inet_ntop(AF_INET, &k->ipv4_info.ip_dst, p2, 32), ntohs(k->tcp_info.dst_port), f1->url_value==NULL?"":" (URL)", f1->packets, f1->bytes);
		}

		flow *newlist_root=NULL;
		for (f1=purge_root; f1!=NULL; f1=f1->next) {
			ipv4_key *k1 = &f1->key;
			unsigned is_in_newlist=0;
			for (f2=newlist_root; f2!=NULL; f2=f2->next) {
				ipv4_key *k2 = &f2->key;
				if (k1->ipv4_info.ip_p == k2->ipv4_info.ip_p && k1->ipv4_info.ip_src.s_addr==k2->ipv4_info.ip_src.s_addr && k1->ipv4_info.ip_dst.s_addr==k2->ipv4_info.ip_dst.s_addr) {
					if ( (k2->tcp_info.src_port==0 && k1->tcp_info.dst_port==k2->tcp_info.dst_port) || (k2->tcp_info.dst_port==0 && k1->tcp_info.src_port==k2->tcp_info.src_port) ){
						f2->bytes+=f1->bytes;
						f2->packets+=f1->packets;
						f2->agg_count++;
						is_in_newlist=1;
						break;
					} else if (k1->tcp_info.src_port!=k2->tcp_info.src_port && k1->tcp_info.dst_port==k2->tcp_info.dst_port ) {
						f2->key.tcp_info.src_port=0;
						f2->bytes+=f1->bytes;
						f2->packets+=f1->packets;
						f2->agg_count++;
						is_in_newlist=2;
						break;
					} else if (k1->tcp_info.src_port==k2->tcp_info.src_port && k1->tcp_info.dst_port!=k2->tcp_info.dst_port ) {
						f2->key.tcp_info.dst_port=0;
						f2->bytes+=f1->bytes;
						f2->packets+=f1->packets;
						f2->agg_count++;
						is_in_newlist=2;
						break;
					}
				} // another proto/src/dst tuple
			}
			if (is_in_newlist==0) {
				f2 = new flow(); mem_total_alloc+=sizeof(flow); mem_c_total_alloc++;
				f2->status = FLOW_STATUS_EXPIRED;
				memcpy(&f2->key, &f1->key, sizeof(struct ipv4_key));
				f2->bytes=f1->bytes; f2->packets=f1->packets;
				f2->url_value=f1->url_value; f1->url_value=NULL;
				f2->url_len = f1->url_len; f1->url_len=0;
				f2->next=newlist_root;
				newlist_root=f2;
			}
		}
		for (f1=purge_root; f1!=NULL; ) {
			f2=f1->next;
			delete f1;
			f1=f2;
		}
		for (f1=newlist_root; f1!=NULL; ) {
			f2=f1->next;
			ipv4_key *k = &f1->key;
			n4debug("\tDUMPNEW %p %02u [%u %s:%u -> %s:%u]%s %lu %llu \n", f1, f1->agg_count, k->ipv4_info.ip_p, inet_ntop(AF_INET, &k->ipv4_info.ip_src, p1, 32), ntohs(k->tcp_info.src_port), inet_ntop(AF_INET, &k->ipv4_info.ip_dst, p2, 32), ntohs(k->tcp_info.dst_port), f1->url_value==NULL?"":" (URL)", f1->packets, f1->bytes);
			(pcap_info->submit_handler)(f1);
			routed_entries++;
			delete f1;
			f1=f2;
		}
	}
}

void flowcache::dump(){
	char p1[32], p2[32];
	flow *f;
	n4debug("FLOWDUMP ENTRIES: %u\n", entries);
	for (int i=0; i<FLOW_CACHE_SIZE; i++) {
		for (f=cache[i]; f!=NULL; f=f->next){
			ipv4_key *k = &f->key;
			unsigned short hash=FLOW_HASH((u_char*)k);
			n4debug("FLOWDUMP  %04x  %p  [%u %s:%u -> %s:%u]%s %lu %llu \n", hash, f, k->ipv4_info.ip_p, inet_ntop(AF_INET, &k->ipv4_info.ip_src, p1, 32), ntohs(k->tcp_info.src_port), inet_ntop(AF_INET, &k->ipv4_info.ip_dst, p2, 32), ntohs(k->tcp_info.dst_port), f->url_value==NULL?"":" (URL)", f->packets, f->bytes);
		}
	}
}

flow *flowcache::find_rev_flow(flow *original){
	ipv4_key *okey=&original->key;
	ipv4_key *key = (ipv4_key *)malloc(sizeof (ipv4_key));
	mem_total_alloc+=sizeof (ipv4_key); mem_c_total_alloc++;
	memset(key, '\0', sizeof (ipv4_key));

	key->ipv4_info.ip_src=okey->ipv4_info.ip_dst;
	key->ipv4_info.ip_dst=okey->ipv4_info.ip_src;
	key->ipv4_info.ip_p=okey->ipv4_info.ip_p;
	key->ipv4_info.ip_tos=okey->ipv4_info.ip_tos;
	key->tcp_info.src_port=okey->tcp_info.dst_port;
	key->tcp_info.dst_port=okey->tcp_info.src_port;


	unsigned short hash=FLOW_HASH((u_char*)key);
	flow *f = cache[hash];
	char p1[32], p2[32];
	unsigned short ohash=FLOW_HASH((u_char*)okey);

	n4debug("find_rev_flow ok=%p, okh=%04x, k=%p, kh=%04x, ofkh=%04x \n", okey, ohash, key, hash, original->hash);
	n4debug("find_rev_flow [%u %s:%u -> %s:%u] flow=%p\n", key->ipv4_info.ip_p, inet_ntop(AF_INET, &key->ipv4_info.ip_src, p1, 32), ntohs(key->tcp_info.src_port), inet_ntop(AF_INET, &key->ipv4_info.ip_dst, p2, 32), ntohs(key->tcp_info.dst_port), f);

	flow *to_ret=NULL;
	for (; f!=NULL; f=f->next){
		if (!memcmp(&f->key, key, sizeof(struct ipv4_key))) {
			to_ret = f;
			break;
		}
	}
	free(key); mem_total_free+=sizeof(ipv4_key); mem_c_total_free++;
	return to_ret;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

