/*************************************************************************
*** This is NeTAMS, version 4
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2011 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id$ */

#include "config.h"
#include "netflow.h"
#include <sys/time.h>
#include <limits.h>
#include <event.h>

#if defined MACOS
#define EVENT_NOPOLL
#define EVENT_NOKQUEUE
#endif

/* global starter variables */
const char *revision="$Rev: 951 $";
long number_of_flows;
double flow_rate;
int verbose=0;

unsigned long long	mem_total_alloc=0;
unsigned long long	mem_url_alloc=0;
unsigned long long	mem_total_free=0;
unsigned long long	mem_url_free=0;

unsigned long long	mem_c_total_alloc=0;
unsigned long long	mem_c_url_alloc=0;
unsigned long long	mem_c_total_free=0;
unsigned long long	mem_c_url_free=0;

char *destination=NULL;
const char *dst_host=NULL;
unsigned short dst_port=20001;

IPStat5Msg *message;
time_t t_now;

int udp_socket_fd;
struct sockaddr_in udp_dest;

long our_addr;
int our_ips;
int randomseed;

unsigned long long total_packets;
unsigned long long total_bytes;

long total_flow_pkts;
long total_flow_rec;

struct timeval expire_tv;

struct clientfd {
    int fd;
    struct event ev;
} client[2];

void show_stats();
void stdin_callback (int fd, short event, void *arg);
void flow_callback (int fd, short event, void *arg);

void signal_handler(int signal){
	if(signal==SIGHUP || signal==SIGTERM || signal==SIGINT) {
		show_stats();
		exit(0);
	}
}

void debug(const char *msg,...){
#ifdef DEBUG
	va_list args;
	char *params;
	if (verbose==0) return;
	params=(char *)malloc(2049);
	va_start(args, msg);
	vsprintf(params, msg, args);
	va_end(args);

	fprintf(stdout, "[%lu]: %s", time(NULL), params);
	fflush(stdout);

	free(params);
#endif
}

int main(int argc, char *argv[]){
	if (argc==1){
		printf("nfgen (NetFlow stream GENerator) for NETAMS 4.0 billing system, %s \n", revision);
		printf("Usage: nfgen <parameters>\n");
		printf("\t-v be verbose\n");
		printf("\t-n number of flows to send\n");
		printf("\t-r flow rate, per second\n");
		printf("\t-d collector destination address:port\n");
		printf("\t-a our subnet IP \n");
		printf("\t-s our subnet size (addresses)\n");
		printf("\t-i random generator seed\n");
		printf("Example: ./nfgen -n 100 -r 1 -i 111 -d 127.0.0.1:20001 -a 192.168.0.0 -s 255\n");
		printf("\n");
		exit(-1);
	}

	number_of_flows=100;
	flow_rate=1;
	our_ips=255;
	randomseed=111;
	// command-line parameters
	int op;
	while((op=getopt(argc, argv, "vn:r:d:a:s:i:")) != EOF){
		switch(op){
			case 'v':
				verbose=1;
				break;
			case 'n':
				number_of_flows = strtol(optarg, NULL, 10);
				break;
			case 'r':
				flow_rate = strtof(optarg, NULL);
				break;
			case 'd':
				destination=(char *)malloc(strlen(optarg)+1);
				strcpy(destination, optarg);
				break;
			case 'a':
				our_addr = ntohl(inet_addr(optarg));
				break;
			case 's':
				our_ips=strtol(optarg, NULL, 10);
				break;
			case 'i':
				randomseed=strtol(optarg, NULL, 10);
				break;
			default:
				break;
		}
	}

	srandom(randomseed);

	if (!destination) dst_host="127.0.0.1";
	else {
		char *c; c=strchr(destination, ':');
		if (c) { dst_port=strtol(c+1, NULL, 10); c[0]='\0'; dst_host=destination; }
		else dst_host=destination;
	}

	setvbuf(stdout, NULL, _IOLBF, 0);

	// handle signals
	signal(SIGCHLD, (sig_t)signal_handler);
	signal(SIGPIPE, (sig_t)signal_handler);
	signal(SIGHUP,  (sig_t)signal_handler);
	signal(SIGTERM, (sig_t)signal_handler);
	signal(SIGINT,  (sig_t)signal_handler);
	signal(SIGUSR1, (sig_t)signal_handler);

	memset(&udp_dest, '\0',  sizeof(udp_dest));
    udp_dest.sin_family=AF_INET;
    udp_dest.sin_port=htons(dst_port);
    inet_aton(dst_host, &udp_dest.sin_addr);
    udp_socket_fd=socket(AF_INET, SOCK_DGRAM, 0);
    if (udp_socket_fd==-1) 	{
       	fprintf(stderr, "failed to create UDP socket: %s\n", strerror(errno));
       	exit(5);
    }

	int status=bigsockbuf(udp_socket_fd, SO_SNDBUF, FT_SO_RCV_BUFSIZE);
	if (status<0) {
		fprintf(stderr, "failed to setsockopt send buffer: %s", strerror(errno));
		exit(5);
	}

	t_now=time(NULL);

	total_flow_pkts = 0;
	total_flow_rec = 0;

	message=(IPStat5Msg*)malloc(sizeof(IPStat5Msg));
	message->header.version=htons(FLOW_VERSION_5);
	message->header.count=0;
	message->header.unix_secs=t_now; //see PR 65
	message->header.sysUpTime=0; //see PR 65
	message->header.engine_id=0;
	message->header.engine_type=htons(31);
	message->header.flow_sequence=htonl(0);

	event_init();

	client[0].fd=STDIN_FILENO; // for DS loop
	client[1].fd=-1; //for flow fwd event

	event_set(&client[0].ev, client[0].fd, EV_READ|EV_PERSIST, stdin_callback, NULL);
	event_add(&client[0].ev, NULL);

	if (flow_rate==0) {
		fprintf(stderr, "flow rate must be nonzero\n");
		exit(5);
	}
	else if (flow_rate>=1) {
		expire_tv.tv_sec = 0;
		expire_tv.tv_usec = 1000000/flow_rate;
	} else {
		expire_tv.tv_sec = (int)((double)1/flow_rate);
		expire_tv.tv_usec = 0;
	}
	evtimer_set(&client[1].ev, flow_callback, NULL);
	evtimer_add(&client[1].ev, &expire_tv);

	debug("flow_rate=%f, delay: %ld sec, %ld usec\n", flow_rate, expire_tv.tv_sec, expire_tv.tv_usec);

	event_dispatch();

	// program stops
	show_stats();
	return 0;
}


void stdin_callback (int fd, short event, void *arg) {
    int ch; while ((ch = getchar()) != '\n' && ch != EOF);
    show_stats();
}

void show_stats(){
	printf("Flows: %ld/%ld\tPackets:%llu\tBytes:%llu\n", total_flow_pkts, total_flow_rec, total_packets, total_bytes);
}

unsigned long rnd(unsigned long max){
//        return max/2;
	return max*((double)random())/RAND_MAX;
}

void genflow();

void flow_callback (int fd, short event, void *arg){
        while (1) {
                genflow();
                usleep(expire_tv.tv_sec*1000000 + expire_tv.tv_usec);
                }
                

}

void genflow(){
	t_now = time(NULL);

	IPFlow5Stat  *record=&message->records[message->header.count];

	long octets = rnd(1000000);
	int packets = rnd(10000);
	record->dOctets=htonl(octets);  total_bytes+=octets;
	record->dPkts=htonl(packets); total_packets+=packets;
	record->input=0;
	record->output=0;
	record->nexthop.s_addr=INADDR_ANY;

	record->tos=0;
	if (rnd(10)>5) record->prot=6; //tcp
	else record->prot=17; //udp


	record->First=htonl(t_now-1000-1000*rnd(600));   //trick here, see PR 65
	record->Last=htonl(t_now-1000);   //trick here, see PR 65

	record->dst_as=0;
	record->dst_mask=0;

	unsigned long a1, a2;
	if (rnd(10)>5) {
		a1=rnd(ULONG_MAX);
		a2=our_addr+rnd(our_ips);
	} else {
		a2=rnd(ULONG_MAX);
		a1=our_addr+rnd(our_ips);
	}

	record->dstaddr.s_addr=htonl(a1);
	record->dstport=htons(rnd(65535));

	record->src_as=0;
	record->src_mask=0;

	record->srcaddr.s_addr=htonl(a2);
	record->srcport=htons(rnd(65535));

	message->header.count++;
	debug("record %u in packet %lu prepared, %lu\n", message->header.count-1, total_flow_pkts+1, octets);
	if (message->header.count==V5_MAXFLOWS) DoSend();

	total_flow_rec++;

	if (total_flow_rec==number_of_flows) {
		DoSend();
		show_stats();
		exit(0);
	}
	
	// flow_callback(fd, event, arg);
	
	// evtimer_add(&client[1].ev, &expire_tv);
}

void DoSend() {
	if (!message || !message->header.count) return;

   	total_flow_pkts++;
   	debug("sending NetFlow v5 packet with %u flows\n", message->header.count);

	unsigned count=message->header.count;
	message->header.count=htons(message->header.count);
	message->header.unix_secs=htonl(message->header.unix_secs);

	if (sendto(udp_socket_fd, message, sizeof(struct Flow5StatHdr)+count*sizeof(struct IPFlow5Stat), 0, (struct sockaddr *)&udp_dest, sizeof(udp_dest)) <0)
        	debug("UDP send failed[%u]: %s", errno, strerror(errno));

	message->header.unix_secs=t_now;
	message->header.count=0;
	message->header.flow_sequence=htonl(total_flow_rec);
}


